import typing

import numpy


def lr_cityblock_cluster_cardinality(ratios: numpy.array) -> float:
    return float(numpy.dot(numpy.asarray(
        a=[1.228800e-01, 1.227127e-01, 7.808845e-02, 4.154137e-02, 5.657729e-02, 3.525646e-02],
        dtype=float,
    ), ratios))


def dt_cityblock_cluster_cardinality(ratios: numpy.array) -> float:
    cardinality, radius, lfd, cardinality_ema, radius_ema, lfd_ema = tuple(ratios)
    if radius <= 1.391490e-02:
        if lfd_ema <= 6.940212e-01:
            if radius <= 1.341821e-02:
                return 5.454838e-01
            else:
                return 3.113860e-01
        else:
            if cardinality <= 2.564081e-02:
                return 3.508578e-01
            else:
                return 7.560313e-01
    else:
        if cardinality_ema <= 3.965033e-01:
            if cardinality <= 1.411780e-01:
                return 8.795414e-01
            else:
                return 9.287172e-01
        else:
            if cardinality_ema <= 7.348774e-01:
                return 9.599685e-01
            else:
                return 9.889817e-01


def lr_cityblock_component_cardinality(ratios: numpy.array) -> float:
    return float(numpy.dot(numpy.asarray(
        a=[1.429870e-02, -1.323484e-02, -1.150261e-02, 3.896381e-02, 4.082664e-02, -1.364604e-02],
        dtype=float,
    ), ratios))


def dt_cityblock_component_cardinality(ratios: numpy.array) -> float:
    cardinality, radius, lfd, cardinality_ema, radius_ema, lfd_ema = tuple(ratios)
    if radius <= 5.148810e-03:
        return 7.500000e-01
    else:
        if cardinality_ema <= 6.438965e-01:
            if lfd_ema <= 9.120842e-01:
                return 9.709770e-01
            else:
                return 9.303896e-01
        else:
            if lfd_ema <= 8.965069e-01:
                return 9.982470e-01
            else:
                return 9.858773e-01


def lr_cityblock_graph_neighborhood(ratios: numpy.array) -> float:
    return float(numpy.dot(numpy.asarray(
        a=[8.262898e-03, 1.537685e-02, 9.422306e-03, 3.740549e-02, 3.891843e-02, -1.250707e-02],
        dtype=float,
    ), ratios))


def dt_cityblock_graph_neighborhood(ratios: numpy.array) -> float:
    cardinality, radius, lfd, cardinality_ema, radius_ema, lfd_ema = tuple(ratios)
    if radius <= 5.148810e-03:
        if cardinality <= 2.540494e-02:
            return 4.437313e-01
        else:
            return 7.500000e-01
    else:
        if cardinality_ema <= 6.354841e-01:
            if cardinality_ema <= 6.253555e-01:
                return 9.625763e-01
            else:
                return 8.007593e-01
        else:
            if cardinality_ema <= 8.838843e-01:
                return 9.887261e-01
            else:
                return 9.979190e-01


def lr_cityblock_parent_cardinality(ratios: numpy.array) -> float:
    return float(numpy.dot(numpy.asarray(
        a=[8.433946e-02, 6.050625e-02, 5.882554e-02, 3.593437e-03, 4.128473e-02, -4.736846e-02],
        dtype=float,
    ), ratios))


def dt_cityblock_parent_cardinality(ratios: numpy.array) -> float:
    cardinality, radius, lfd, cardinality_ema, radius_ema, lfd_ema = tuple(ratios)
    if radius <= 1.472459e-02:
        if cardinality <= 2.505734e-02:
            if radius <= 1.655350e-04:
                return 3.151504e-01
            else:
                return 6.447252e-01
        else:
            if cardinality <= 1.076636e-01:
                return 7.414646e-01
            else:
                return 9.229898e-01
    else:
        if cardinality <= 5.942344e-01:
            if lfd_ema <= 9.527803e-01:
                return 9.468785e-01
            else:
                return 8.911853e-01
        else:
            if cardinality_ema <= 9.790418e-01:
                return 9.705729e-01
            else:
                return 9.989975e-01


def lr_cityblock_stationary_probabilities(ratios: numpy.array) -> float:
    return float(numpy.dot(numpy.asarray(
        a=[4.659433e-02, -5.014006e-02, -6.017402e-02, 5.812719e-02, 1.466290e-01, 1.266893e-03],
        dtype=float,
    ), ratios))


def dt_cityblock_stationary_probabilities(ratios: numpy.array) -> float:
    cardinality, radius, lfd, cardinality_ema, radius_ema, lfd_ema = tuple(ratios)
    if radius <= 2.358828e-04:
        if radius_ema <= 1.177303e-02:
            if radius_ema <= 8.709679e-05:
                return 1.400713e-01
            else:
                return 1.698153e-01
        else:
            return 4.245149e-01
    else:
        if cardinality_ema <= 1.807206e-02:
            if radius_ema <= 7.669807e-01:
                return 5.307418e-01
            else:
                return 9.984011e-01
        else:
            if cardinality_ema <= 4.135659e-01:
                return 9.372930e-01
            else:
                return 9.907336e-01


def lr_cityblock_vertex_degree(ratios: numpy.array) -> float:
    return float(numpy.dot(numpy.asarray(
        a=[1.032663e-01, 1.232432e-01, 7.317461e-02, 1.084027e-02, 9.541312e-02, 1.760110e-02],
        dtype=float,
    ), ratios))


def dt_cityblock_vertex_degree(ratios: numpy.array) -> float:
    cardinality, radius, lfd, cardinality_ema, radius_ema, lfd_ema = tuple(ratios)
    if radius <= 1.391490e-02:
        if lfd_ema <= 6.932603e-01:
            if cardinality <= 1.076636e-01:
                return 4.198158e-01
            else:
                return 8.463385e-01
        else:
            if cardinality <= 2.564081e-02:
                return 3.269299e-01
            else:
                return 7.921413e-01
    else:
        if cardinality_ema <= 7.950329e-01:
            if cardinality <= 4.997393e-01:
                return 9.251187e-01
            else:
                return 9.570666e-01
        else:
            if radius_ema <= 4.155880e-01:
                return 9.516691e-01
            else:
                return 9.928017e-01


def lr_euclidean_cluster_cardinality(ratios: numpy.array) -> float:
    return float(numpy.dot(numpy.asarray(
        a=[1.313924e-01, 1.326884e-01, 9.136274e-02, 2.134787e-02, 3.100747e-02, 3.298891e-02],
        dtype=float,
    ), ratios))


def dt_euclidean_cluster_cardinality(ratios: numpy.array) -> float:
    cardinality, radius, lfd, cardinality_ema, radius_ema, lfd_ema = tuple(ratios)
    if radius <= 1.067198e-02:
        if cardinality <= 1.079019e-01:
            if lfd_ema <= 9.999740e-01:
                return 5.265128e-01
            else:
                return 8.834613e-01
        else:
            if cardinality <= 3.330053e-01:
                return 8.695605e-01
            else:
                return 9.694478e-01
    else:
        if cardinality <= 4.997283e-01:
            if lfd <= 2.929797e-01:
                return 8.696701e-01
            else:
                return 9.258335e-01
        else:
            if cardinality_ema <= 7.698584e-01:
                return 9.665650e-01
            else:
                return 9.917949e-01


def lr_euclidean_component_cardinality(ratios: numpy.array) -> float:
    return float(numpy.dot(numpy.asarray(
        a=[5.582536e-02, 2.442987e-02, 8.037801e-03, 1.539072e-02, 3.654952e-02, 1.429881e-02],
        dtype=float,
    ), ratios))


def dt_euclidean_component_cardinality(ratios: numpy.array) -> float:
    cardinality, radius, lfd, cardinality_ema, radius_ema, lfd_ema = tuple(ratios)
    if radius_ema <= 5.111776e-03:
        if radius <= 9.307770e-05:
            return 7.500000e-01
        else:
            return 6.997932e-01
    else:
        if cardinality <= 1.494950e-01:
            if cardinality_ema <= 4.477562e-02:
                return 8.816036e-01
            else:
                return 9.351654e-01
        else:
            if cardinality_ema <= 6.924688e-01:
                return 9.702070e-01
            else:
                return 9.932433e-01


def lr_euclidean_graph_neighborhood(ratios: numpy.array) -> float:
    return float(numpy.dot(numpy.asarray(
        a=[9.585250e-02, 6.025230e-02, 3.753800e-02, -3.118970e-03, 7.559676e-02, 1.875789e-02],
        dtype=float,
    ), ratios))


def dt_euclidean_graph_neighborhood(ratios: numpy.array) -> float:
    cardinality, radius, lfd, cardinality_ema, radius_ema, lfd_ema = tuple(ratios)
    if radius_ema <= 2.359367e-05:
        if lfd <= 3.415343e-01:
            if cardinality <= 4.490425e-02:
                return 1.299296e-01
            else:
                return 1.299296e-01
        else:
            return 4.405503e-01
    else:
        if cardinality_ema <= 1.619149e-02:
            if cardinality_ema <= 3.227095e-03:
                return 5.649648e-01
            else:
                return 7.366586e-01
        else:
            if radius_ema <= 5.773949e-02:
                return 8.322585e-01
            else:
                return 9.700278e-01


def lr_euclidean_parent_cardinality(ratios: numpy.array) -> float:
    return float(numpy.dot(numpy.asarray(
        a=[9.211533e-02, 7.063269e-02, 2.331588e-02, 7.911250e-04, 1.672235e-02, 3.891130e-03],
        dtype=float,
    ), ratios))


def dt_euclidean_parent_cardinality(ratios: numpy.array) -> float:
    cardinality, radius, lfd, cardinality_ema, radius_ema, lfd_ema = tuple(ratios)
    if radius <= 1.067198e-02:
        if cardinality <= 2.828062e-02:
            if radius <= 9.307770e-05:
                return 3.440542e-01
            else:
                return 6.220238e-01
        else:
            if lfd <= 2.580366e-01:
                return 5.624770e-01
            else:
                return 8.176447e-01
    else:
        if cardinality <= 3.860220e-02:
            if cardinality <= 3.826937e-02:
                return 8.636209e-01
            else:
                return 5.281209e-01
        else:
            if cardinality <= 4.997400e-01:
                return 9.494346e-01
            else:
                return 9.803474e-01


def lr_euclidean_stationary_probabilities(ratios: numpy.array) -> float:
    return float(numpy.dot(numpy.asarray(
        a=[8.131170e-02, 1.629200e-02, -4.917042e-02, 3.507954e-02, -1.800446e-03, -5.963697e-03],
        dtype=float,
    ), ratios))


def dt_euclidean_stationary_probabilities(ratios: numpy.array) -> float:
    cardinality, radius, lfd, cardinality_ema, radius_ema, lfd_ema = tuple(ratios)
    if radius_ema <= 3.118177e-02:
        if lfd <= 8.090920e-01:
            if lfd_ema <= 2.858897e-01:
                return 4.526594e-01
            else:
                return 8.478856e-01
        else:
            if lfd_ema <= 6.024930e-01:
                return 5.471349e-03
            else:
                return 2.958559e-01
    else:
        if cardinality <= 1.494950e-01:
            if radius_ema <= 9.417503e-01:
                return 9.296579e-01
            else:
                return 7.623177e-01
        else:
            if cardinality_ema <= 1.749870e-02:
                return 7.646179e-01
            else:
                return 9.863799e-01


def lr_euclidean_vertex_degree(ratios: numpy.array) -> float:
    return float(numpy.dot(numpy.asarray(
        a=[1.096380e-01, 1.658747e-01, 1.045492e-01, -1.207694e-03, 6.186889e-02, 2.726535e-02],
        dtype=float,
    ), ratios))


def dt_euclidean_vertex_degree(ratios: numpy.array) -> float:
    cardinality, radius, lfd, cardinality_ema, radius_ema, lfd_ema = tuple(ratios)
    if radius <= 1.067198e-02:
        if lfd <= 9.634169e-01:
            if lfd_ema <= 9.999111e-01:
                return 4.713993e-01
            else:
                return 9.783193e-01
        else:
            if radius <= 1.020328e-02:
                return 9.089947e-01
            else:
                return 4.464702e-01
    else:
        if cardinality <= 4.997769e-01:
            if lfd <= 9.999996e-01:
                return 9.335233e-01
            else:
                return 7.793101e-01
        else:
            if radius_ema <= 7.430525e-01:
                return 9.627948e-01
            else:
                return 9.921575e-01


META_MODELS = {
    'lr_cityblock_cluster_cardinality': lr_cityblock_cluster_cardinality,
    'dt_cityblock_cluster_cardinality': dt_cityblock_cluster_cardinality,
    'lr_cityblock_component_cardinality': lr_cityblock_component_cardinality,
    'dt_cityblock_component_cardinality': dt_cityblock_component_cardinality,
    'lr_cityblock_graph_neighborhood': lr_cityblock_graph_neighborhood,
    'dt_cityblock_graph_neighborhood': dt_cityblock_graph_neighborhood,
    'lr_cityblock_parent_cardinality': lr_cityblock_parent_cardinality,
    'dt_cityblock_parent_cardinality': dt_cityblock_parent_cardinality,
    'lr_cityblock_stationary_probabilities': lr_cityblock_stationary_probabilities,
    'dt_cityblock_stationary_probabilities': dt_cityblock_stationary_probabilities,
    'lr_cityblock_vertex_degree': lr_cityblock_vertex_degree,
    'dt_cityblock_vertex_degree': dt_cityblock_vertex_degree,
    'lr_euclidean_cluster_cardinality': lr_euclidean_cluster_cardinality,
    'dt_euclidean_cluster_cardinality': dt_euclidean_cluster_cardinality,
    'lr_euclidean_component_cardinality': lr_euclidean_component_cardinality,
    'dt_euclidean_component_cardinality': dt_euclidean_component_cardinality,
    'lr_euclidean_graph_neighborhood': lr_euclidean_graph_neighborhood,
    'dt_euclidean_graph_neighborhood': dt_euclidean_graph_neighborhood,
    'lr_euclidean_parent_cardinality': lr_euclidean_parent_cardinality,
    'dt_euclidean_parent_cardinality': dt_euclidean_parent_cardinality,
    'lr_euclidean_stationary_probabilities': lr_euclidean_stationary_probabilities,
    'dt_euclidean_stationary_probabilities': dt_euclidean_stationary_probabilities,
    'lr_euclidean_vertex_degree': lr_euclidean_vertex_degree,
    'dt_euclidean_vertex_degree': dt_euclidean_vertex_degree,
}

META_ML_MODELS: typing.List[typing.Tuple[str, str, typing.Callable[[numpy.array], float]]] = [
    # tuple of (metric, method, function)
    (name.split('_')[1], '_'.join(name.split('_')[2:]), method)
    for name, method in META_MODELS.items()
]



import json
import logging
import pathlib
import subprocess
import typing

import h5py
import numpy
import scipy.io

from ..utils import constants
from ..utils import helpers

__all__ = ['DATASET_URLS', 'ChaodaData']

logger = logging.getLogger(__name__)
logger.setLevel(constants.LOG_LEVEL)

MAT_FILE_URLS: dict[str, str] = {
    'annthyroid': 'https://www.dropbox.com/s/aifk51owxbogwav/annthyroid.mat?dl=0',
    'arrhythmia': 'https://www.dropbox.com/s/lmlwuspn1sey48r/arrhythmia.mat?dl=0',
    'breastw': 'https://www.dropbox.com/s/g3hlnucj71kfvq4/breastw.mat?dl=0',
    'cardio': 'https://www.dropbox.com/s/galg3ihvxklf0qi/cardio.mat?dl=0',
    'cover': 'https://www.dropbox.com/s/awx8iuzbu8dkxf1/cover.mat?dl=0',
    'glass': 'https://www.dropbox.com/s/iq3hjxw77gpbl7u/glass.mat?dl=0',
    'ionosphere': 'https://www.dropbox.com/s/lpn4z73fico4uup/ionosphere.mat?dl=0',
    'lympho': 'https://www.dropbox.com/s/ag469ssk0lmctco/lympho.mat?dl=0',
    'mammography': 'https://www.dropbox.com/s/tq2v4hhwyv17hlk/mammography.mat?dl=0',
    'mnist': 'https://www.dropbox.com/s/n3wurjt8v9qi6nc/mnist.mat?dl=0',
    'musk': 'https://www.dropbox.com/s/we6aqhb0m38i60t/musk.mat?dl=0',
    'optdigits': 'https://www.dropbox.com/s/w52ndgz5k75s514/optdigits.mat?dl=0',
    'pendigits': 'https://www.dropbox.com/s/1x8rzb4a0lia6t1/pendigits.mat?dl=0',
    'pima': 'https://www.dropbox.com/s/mvlwu7p0nyk2a2r/pima.mat?dl=0',
    'satellite': 'https://www.dropbox.com/s/dpzxp8jyr9h93k5/satellite.mat?dl=0',
    'satimage-2': 'https://www.dropbox.com/s/hckgvu9m6fs441p/satimage-2.mat?dl=0',
    'shuttle': 'https://www.dropbox.com/s/mk8ozgisimfn3dw/shuttle.mat?dl=0',
    'thyroid': 'https://www.dropbox.com/s/bih0e15a0fukftb/thyroid.mat?dl=0',
    'vertebral': 'https://www.dropbox.com/s/5kuqb387sgvwmrb/vertebral.mat?dl=0',
    'vowels': 'https://www.dropbox.com/s/pa26odoq6atq9vx/vowels.mat?dl=0',
    'wbc': 'https://www.dropbox.com/s/ebz9v9kdnvykzcb/wbc.mat?dl=0',
    'wine': 'https://www.dropbox.com/s/uvjaudt2uto7zal/wine.mat?dl=0',
}

HDF5_FILE_URLS = {
    'http': 'https://www.dropbox.com/s/iy9ucsifal754tp/http.mat?dl=0',
    'smtp': 'https://www.dropbox.com/s/dbv2u4830xri7og/smtp.mat?dl=0',
}

DATASET_URLS = {**MAT_FILE_URLS, **HDF5_FILE_URLS}

MMapMode = typing.Literal["r", None]


class ChaodaData:

    __slots__ = [
        'data_dir',
        'name',
        'url',
        'mmap_threshold',
        'mmap_mode',
        'raw_path',
        'features_path',
        'normalized_features_path',
        'scores_path',
    ]

    def __init__(self, data_dir: pathlib.Path, name: str, url: str, mmap_threshold: int = 1024 ** 3, mmap_mode: MMapMode = None):
        self.data_dir = data_dir
        self.name = name
        self.url = url
        self.mmap_threshold = mmap_threshold
        self.mmap_mode: MMapMode = mmap_mode

        data_dir.mkdir(exist_ok=True)
        raw_dir = data_dir.joinpath('raw')
        raw_dir.mkdir(exist_ok=True)

        preprocessed_dir = data_dir.joinpath('preprocessed')
        preprocessed_dir.mkdir(exist_ok=True)

        self.raw_path = raw_dir.joinpath(f'{name}.mat')
        self.features_path = preprocessed_dir.joinpath(f'{name}_features.npy')
        self.normalized_features_path = preprocessed_dir.joinpath(f'{name}_features_normalized.npy')
        self.scores_path = preprocessed_dir.joinpath(f'{name}_scores.npy')

    def save(self) -> pathlib.Path:
        save_path = self.data_dir.joinpath('classes').joinpath(f'{self.name}.json')
        save_path.parent.mkdir(exist_ok=True)
        attributes = {
            'name': self.name,
            'url': self.url,
            'mmap_threshold': self.mmap_threshold,
            'mmap_mode': self.mmap_mode,
            'raw_path': self.raw_path,
            'features_path': self.features_path,
            'normalized_features_path': self.normalized_features_path,
            'scores_path': self.scores_path,
        }

        for k, v in attributes.items():
            if isinstance(v, pathlib.Path):
                relative_path = v.relative_to(self.data_dir)
                attributes[k] = str(relative_path)

        with open(save_path, 'w') as writer:
            json.dump(attributes, writer, indent=4)
        return save_path

    @staticmethod
    def load(data_dir: pathlib.Path, name: str) -> 'ChaodaData':
        save_path = data_dir.joinpath('classes').joinpath(f'{name}.json')
        with open(save_path, 'r') as reader:
            attributes = json.load(reader)

        data = ChaodaData(
            data_dir=data_dir,
            name=attributes['name'],
            url=attributes['url'],
            mmap_threshold=int(attributes['mmap_threshold']),
            mmap_mode=attributes['mmap_mode'],
        )
        for k in ChaodaData.__slots__:
            if 'path' in k:
                v = data_dir.joinpath(pathlib.Path(attributes[k]))
                setattr(data, k, v)

        return data

    def download(
            self,
            extension: str = 'mat',
            force: bool = False,
            suppress_stdout: bool = True,
            suppress_stderr: bool = True,
    ):
        self.raw_path = self.raw_path.with_name(f'{self.name}.{extension}')

        if not force and self.raw_path.exists():
            return

        self.raw_path.unlink(missing_ok=True)

        kwargs = dict()
        if suppress_stdout:
            kwargs['stdout'] = subprocess.DEVNULL
        if suppress_stderr:
            kwargs['stderr'] = subprocess.DEVNULL
        # wget.download(self.url, out=str(self.raw_path))
        subprocess.run(['wget', self.url, '-O', self.raw_path], **kwargs)

        size = self.raw_path.stat().st_size
        if size < 1024:
            warning = f'Downloaded file for {self.name} was too small. {size} bytes.'
            logging.warning(warning)
            raise BytesWarning(warning)

        if size > self.mmap_threshold:
            self.mmap_mode = 'r'

        return

    def preprocess(self, normalization_mode: str = 'gaussian'):
        helpers.catch_normalization_mode(normalization_mode)

        data_dict = dict()
        if self.name in HDF5_FILE_URLS:
            with h5py.File(self.raw_path, 'r') as reader:
                data_dict['X'] = numpy.asarray(reader['X']).T
                data_dict['y'] = numpy.asarray(reader['y']).T
        else:
            data_dict = scipy.io.loadmat(str(self.raw_path))

        features = numpy.asarray(data_dict['X'], dtype=numpy.float32)
        numpy.save(str(self.features_path), features, allow_pickle=False, fix_imports=False)

        features = helpers.normalize(features, mode=normalization_mode)
        numpy.save(str(self.normalized_features_path), features, allow_pickle=False, fix_imports=False)

        scores = numpy.asarray(data_dict['y'], dtype=numpy.uint8).squeeze()
        numpy.save(str(self.scores_path), scores, allow_pickle=False, fix_imports=False)

        return

    @property
    def features(self):
        if not self.features_path.exists():
            raise ValueError(f'Dataset {self.name} as not yet been downloaded.')
        else:
            return numpy.load(str(self.features_path), self.mmap_mode)

    @property
    def scores(self):
        if not self.features_path.exists():
            raise ValueError(f'Dataset {self.name} as not yet been downloaded.')
        else:
            return numpy.load(str(self.scores_path))



import logging

import numpy

from ..core import Cluster
from ..core import criterion
from ..core import Manifold
from ..core import types
from ..utils import constants

logger = logging.getLogger(__name__)
logger.setLevel(constants.LOG_LEVEL)

ClusterResults = dict[Cluster, float]
Results = dict[int, float]


class CAKES:
    def __init__(self, data: types.Dataset, metric: types.Metric):
        self.data: types.Dataset = data
        self.metric: types.Metric = metric
        self.manifold: Manifold = Manifold(self.data, self.metric)
        self.distance = self.manifold.distance
        self.root: Cluster = self.manifold.root

    @staticmethod
    def from_manifold(manifold: Manifold) -> 'CAKES':
        search = CAKES(manifold.data, manifold.metric)
        search.manifold = manifold
        search.distance = manifold.distance
        search.root = manifold.root
        return search

    @property
    def depth(self) -> int:
        """ returns the depth of the search tree. """
        return self.manifold.depth

    def build(self, *, max_depth: int = None) -> 'CAKES':
        """ Builds the search tree upto leaves, or an optional maximum depth.

        This method can be called repeatedly, with higher depth values, to further increase the depth of the tree.

        :param max_depth: optional maximum depth of search tree
        :return: the modified Search object.
        """
        if max_depth is None:
            logger.info(f'Building CAKES tree to leaves.')
            self.manifold.build(criterion.Layer(-1))
        elif max_depth < 1:
            raise ValueError(f'Expected a positive integer for max_depth. Got {max_depth} instead.')
        elif max_depth > self.depth:
            logger.info(f'Building CAKES tree to depth {max_depth}.')
            self.manifold.build_tree(criterion.MaxDepth(max_depth), criterion.Layer(-1))
        return self

    def rnn(self, query: types.Datum, radius: float, *, max_depth: int = None) -> Results:
        """ Performs rho-nearest neighbors search around query with given radius.

        :param query: point around which to search.
        :param radius: distance from query within which te return all hits.
        :param max_depth: optional maximum search depth.
        :return: dictionary of index-of-hit -> distance-to-hit.
        """
        candidate_clusters: ClusterResults = self.tree_search(query, radius, start=None, max_depth=max_depth)
        return self.leaf_search(query, radius, candidate_clusters)

    def rnn_points(self, query: types.Datum, radius: float, *, max_depth: int = None) -> types.Dataset:
        """ Same as rnn, but returns the raw points rather than a dictionary of indices and distances. """
        results = [(d, p) for p, d in self.rnn(query, radius, max_depth=max_depth).items()]
        return self.data[[p for _, p in sorted(results)]]

    def knn(self, query: types.Datum, k: int, *, max_depth: int = None) -> Results:
        """ Performs k-nearest neighbors search around query.

        :param query: point around which to look.
        :param k: number of closest neighbors to look for.
        :param max_depth: optional maximum search depth.
        :return: dictionary of index-of-neighbor -> distance-to-neighbor.
        """
        if k < 1:
            raise ValueError(f'k must be a positive integer. Got {k}')

        radius: float = (k / self.root.cardinality) * self.root.radius
        assert radius > 0, f'expected a positive value for radius. Got {radius:.2e} instead.'

        hits: Results = self.rnn(query, radius, max_depth=max_depth)
        while len(hits) == 0:  # make sure to have non-zero hits
            radius *= 2
            hits = self.rnn(query, radius, max_depth=max_depth)

        while len(hits) < k:  # make sure to have more at least k hits
            # calculate lfd in ball around query
            if len(hits) == 1:
                lfd = 1
            else:
                half_count = len([point for point, distance in hits.items() if distance <= radius / 2])
                lfd = 1 if half_count == 0 else numpy.log2(len(hits) / half_count)

            # increase radius as guided by lfd
            factor = (k / len(hits)) ** (1 / (lfd + constants.EPSILON))
            assert factor > 1, f'expected factor to be greater than 1. Got {factor:.2e} instead.'
            radius *= factor

            # rerun rnn search
            hits = self.rnn(query, radius, max_depth=max_depth)

        # sort hits in non-decreasing order of distance to query
        results: list[tuple[float, int]] = list(sorted([(distance, point) for point, distance in hits.items()]))
        return {point: distance for distance, point in results[:k]}

    def knn_points(self, query: types.Datum, k: int, *, max_depth: int = None) -> types.Dataset:
        """ Same as knn, but returns the raw points rather than a dictionary of indices and distances. """
        results = [(d, p) for p, d in self.knn(query, k, max_depth=max_depth).items()]
        return self.data[[p for _, p in sorted(results)]]

    def _check_shape(self, points: numpy.ndarray) -> None:
        if len(points.shape) != len(self.data.shape):
            raise ValueError(f'wrong number of dimensions in shape. Expected {len(self.data.shape)}, but got {len(points.shape)}')
        elif points.shape[1:] != self.data.shape[1:]:
            raise ValueError(f'wrong dimensionality of points. Expected {self.data.shape[1:]}, but got {points.shape[1:]}')
        else:
            return

    def _parse_query(self, query: types.Datum) -> numpy.ndarray:
        if isinstance(query, int):
            query: numpy.ndarray = numpy.asarray([self.data[query]])
        query = numpy.expand_dims(query, 0)
        self._check_shape(query)
        return query

    def tree_search(self, query: types.Datum, radius: float, *, start: Cluster = None, max_depth: int = None) -> ClusterResults:
        """ Performs tree-search for the query, starting at the given cluster.

        Consider the sphere around 'query' with the given 'radius'.
        This method performs a breadth-first search over the tree and, at each iteration,
        discards the clusters that do not have overlap with this sphere.
        Once the search terminates, the remaining clusters are returned.

        :param query: point around which to search.
        :param radius: distance from point within which to look.
        :param start: starting cluster for search.
        :param max_depth: optional maximum search depth starting at Cluster.
        :return: dictionary of cluster -> distance between query and cluster center
        """
        return self.tree_search_history(query, radius, start=start, max_depth=max_depth)[1]

    def tree_search_history(
            self,
            query: types.Datum,
            radius: float,
            *,
            start: Cluster = None,
            max_depth: int = None,
    ) -> tuple[ClusterResults, ClusterResults]:
        """ Same as tree-search, except that it also returns the history of candidate clusters at each depth. """
        start = start or self.root

        if max_depth is None:
            max_depth = self.manifold.depth
        elif max_depth < 0:
            raise ValueError(f'max_depth must be a non-negative integer. Got {max_depth} instead.')
        else:
            max_depth = min(self.manifold.depth, start.depth + max_depth)

        query = self._parse_query(query)
        candidates: ClusterResults = {start: -1}
        history: ClusterResults = dict()
        hits: ClusterResults = dict()
        for depth in range(start.depth, max_depth + 1):
            logger.debug(f'Searching tree at {depth = }')

            if len(candidates) == 0:
                break
            # find distances to centers of candidate clusters
            clusters: list[Cluster] = list(candidates.keys())
            centers: numpy.ndarray = numpy.asarray([cluster.center for cluster in clusters])
            self._check_shape(centers)
            distances: list[float] = list(self.distance(query, centers)[0])

            candidates = {  # filter out clusters that are too far away
                cluster: distance for cluster, distance in zip(clusters, distances)
                if distance <= (radius + cluster.radius)
            }
            subsumed: set[Cluster] = {  # set of clusters that lie completely within radius of query
                cluster for cluster, distance in candidates.items()
                if (distance + cluster.radius) <= radius
            }
            hits.update({  # add subsumed clusters to hits, singletons should already be included here
                cluster: distance
                for cluster, distance in candidates.items()
                if (cluster in subsumed) or (len(cluster.children) == 0)
            })

            history.update(candidates)  # add candidates from this depth to history

            candidates = {  # expand search to children
                child: -1
                for cluster in candidates.keys()
                for child in cluster.children
                if cluster not in hits
            }

        logger.debug(f'Found {len(hits)} candidate clusters from tree-search.')
        return history, hits

    def leaf_search(self, query: types.Datum, radius: float, clusters: ClusterResults) -> Results:
        """ Performs leaf search for query on given clusters.

        :param query: point around which to look.
        :param radius: distance from point within which to look.
        :param clusters: candidate clusters which need to be exhaustively searched.
        :return: dictionary of index-of-hit -> distance-to-hit.
        """
        query = self._parse_query(query)

        # get indices of all points in candidate clusters
        argpoints: list[int] = [point for cluster in clusters for point in cluster.argpoints]
        logger.debug(f'Performing leaf-search over {len(argpoints)} points.')

        # get distances from query to candidate points
        points: numpy.ndarray = self.data[argpoints]
        self._check_shape(points)
        distances: list[float] = list(self.distance(query, points)[0])

        # Filter hits with points within radius of query
        hits = {point: distance for point, distance in zip(argpoints, distances) if distance <= radius}
        return hits



import typing

import numpy

Dataset = numpy.ndarray  # the entire dataset, TODO: Replace with custom DataLoader class
Datum = numpy.ndarray  # a point,
Distance = typing.Union[numpy.float64, float]
Indices = list[int]  # TODO: Rename to Indices of Argpoints
DistanceFunc = typing.Callable[[Datum, Datum], Distance]
Metric = typing.Union[str, DistanceFunc]  # TODO: Compile a list of metrics supported by scipy and by CLAM.



""" This module provides the CHAODA algorithms implemented on top of CLAM.
"""
import logging
import typing

import numpy

from .meta_ml_models import META_ML_MODELS
from ..core import Cluster
from ..core import criterion
from ..core import Graph
from ..core import Manifold
from ..core import types
from ..search import CAKES
from ..utils import constants
from ..utils import helpers

logger = logging.getLogger(__name__)
logger.setLevel(constants.LOG_LEVEL)

ClusterScores = dict[Cluster, float]
Scores = dict[int, float]
VOTING_MODES = [
    'mean',
    'product',
    'median',
    'min',
    'max',
    'p25',
    'p75',
]


class CHAODA:
    method_names = [
        'cluster_cardinality',
        'component_cardinality',
        'graph_neighborhood',
        'parent_cardinality',
        'stationary_probabilities',
        'vertex_degree',
    ]
    """ This class provides implementations of the CHAODA algorithms on top of the CLAM framework.
    """
    # TODO: Allow weights for each method in ensemble.
    # TODO: Allow meta-ml with users' own datasets.
    # TODO: Look at DSD (diffusion-state-distance) as a possible individual-method.
    # noinspection PyTypeChecker
    def __init__(
            self, *,
            metrics: list[types.Metric] = None,
            min_depth: int = 4,
            max_depth: int = 25,
            min_points: int = 1,
            meta_ml_functions: list[tuple[str, str, typing.Callable[[numpy.array], float]]] = None,
            normalization_mode: typing.Optional[str] = 'gaussian',
            speed_threshold: typing.Optional[int] = 128,
    ):
        """ Creates and initializes a CHAODA object.

        :param metrics: A list of distance metrics to use for creating manifolds.
                        A metric must deterministically produce positive real numbers for each pair of instances.
                        A metric need not obey the triangle inequality.
                        Any such metrics allowed by scipy are allowed here, as are user-defined functions (so long as scipy accepts them).
        :param min_depth: The minimum depth clusters that can be selected for optimal graphs using meta-ml.
        :param max_depth: The max-depth of the cluster-trees in the manifolds.
        :param min_points: The minimum number of points in a cluster before it can be partitioned further.
        :param meta_ml_functions: A list of tuples of (metric-name, method-name, meta-ml-ranking-function).
        :param normalization_mode: What normalization mode to use. Must be one of 'linear', 'gaussian', or 'sigmoid'.
        :param speed_threshold: number of clusters above which to skip the slow methods.
        """
        self.metrics: list[types.Metric] = metrics or ['euclidean', 'cityblock']

        # Set criteria for building manifolds
        self.max_depth: int = max_depth
        self.min_points: int = min_points

        # Set meta-ml selection criteria
        self.min_depth: int = min_depth

        meta_ml_functions = META_ML_MODELS if meta_ml_functions is None else meta_ml_functions

        metrics = {metric for metric, _, _ in meta_ml_functions}
        if not metrics <= set(self.metrics):
            raise ValueError(f'some meta-ml-functions reference metrics not being used by CHAODA. {set(self.metrics) - metrics}')
        methods = {method for _, method, _ in meta_ml_functions}
        if not methods <= set(self.method_names):
            raise ValueError(f'some meta-ml-functions reference methods not being used by CHAODA. {set(self.method_names) - methods}')

        self._criteria = {
            metric: {method: list() for method in self.method_names}
            for metric in self.metrics
        }
        for metric, method, function in meta_ml_functions:
            self._criteria[metric][method].append(criterion.MetaMLSelect(function, self.min_depth))

        self.speed_threshold: typing.Optional[int] = speed_threshold

        if normalization_mode is not None:
            helpers.catch_normalization_mode(normalization_mode)
        self.normalization_mode: typing.Optional[str] = normalization_mode

        # dict of name -> individual-method
        self._names: dict[str, typing.Callable[[Graph], ClusterScores]] = {
            'cluster_cardinality': self._cluster_cardinality,
            'component_cardinality': self._component_cardinality,
            'graph_neighborhood': self._graph_neighborhood,
            'parent_cardinality': self._parent_cardinality,
            'stationary_probabilities': self._stationary_probabilities,
            'vertex_degree': self._vertex_degree,
        }
        self.slow_methods: list[str] = [
            'component_cardinality',
            'graph_neighborhood',
            'stationary_probabilities',
        ]

        # Values to be set later

        self._voting_mode: str = None

        # list of manifolds build during the fit method.
        self._manifolds: list[Manifold] = None

        # These graphs are used with the individual-methods.
        #  This is a list of tuples os (method-name, graph).
        self._graphs: list[tuple[str, Graph]] = None

        # Each item in the list corresponds to a Graph and is a dictionary of
        # the outlier-scores for the Clusters in that Graph.
        self._cluster_scores: list[ClusterScores] = None

        # A list of all outlier-scores arrays to be voted amongst.
        self._individual_scores: list[numpy.array] = None

        # outlier-scores for each point in the fitted data, after voting.
        self._scores: numpy.array = None

    @property
    def manifolds(self) -> list[Manifold]:
        """ Returns the list of manifolds being used for anomaly-detection. """
        return self._manifolds

    @property
    def scores(self) -> numpy.ndarray:
        """ Returns the scores for data on which the model was last fit. """
        if self._scores is None:
            raise ValueError(f'Scores are currently empty. Please call the fit method.')
        return self._scores

    @property
    def individual_scores(self) -> list[numpy.ndarray]:
        if self._individual_scores is None:
            raise ValueError(f'Individual-scores are currently empty. Please call the fit method.')
        return self._individual_scores

    def build_manifolds(self, data: numpy.array, indices: list[int] = None) -> list[Manifold]:
        """ Builds the list of manifolds for the class.

        :param data: numpy array of data where the rows are instances and the columns are features.
        :param indices: Optional. List of indexes of data to which to restrict the Manifolds.
        :return: The list of manifolds.
        """
        indices = list(range(data.shape[0])) if indices is None else indices
        criteria: list[criterion.Criterion] = [criterion.MaxDepth(self.max_depth), criterion.MinPoints(self.min_points)]
        self._manifolds = [Manifold(data, metric, indices).build(*criteria) for metric in self.metrics]
        return self._manifolds

    def fit(self, data: numpy.array, *, indices: list[int] = None, voting: str = 'mean') -> 'CHAODA':
        """ Fits the anomaly detector to the data.

        :param data: numpy array of data where the rows are instances and the columns are features.
        :param indices: Optional. List of indexes of data to which to restrict CHAODA.
        :param voting: How to vote among scores.
        :return: the fitted CHAODA model.
        """
        if voting not in VOTING_MODES:
            raise ValueError(f'\'voting\' must be one of {VOTING_MODES}. Got {voting} instead.')
        self._voting_mode = voting

        self.build_manifolds(data, indices)
        self._graphs = list()
        for manifold in self._manifolds:
            for method, meta_ml_criteria in self._criteria[manifold.metric].items():
                manifold.add_graphs(*meta_ml_criteria)
                for graph in manifold.graphs[-len(meta_ml_criteria):]:
                    if method in self.slow_methods and graph.cardinality > self.speed_threshold:
                        continue
                    self._graphs.append((method, graph))
        self._ensemble()
        return self

    def predict_single(self, query: numpy.array) -> float:
        """ Predict the anomaly score for a single query.
        """
        scores = list()
        for cluster_scores in self._cluster_scores:
            manifold = next(iter(cluster_scores.keys())).manifold
            searcher = CAKES.from_manifold(manifold)
            hits = list(searcher.tree_search_history(query, radius=0)[0].keys())
            intersection = [cluster for cluster in hits if cluster in cluster_scores]
            if len(intersection) > 0:
                individual_scores = [cluster_scores[cluster] for cluster in intersection]
            else:
                individual_scores = [1.]
            scores.append(self._vote(individual_scores))
        score = self._vote(scores)
        return score

    def predict(self, queries: numpy.array) -> numpy.array:
        """ Predict the anomaly score for a 2d array of queries.
        """
        scores = list()
        for i in range(queries.shape[0]):
            logger.info(f'Predicting anomaly score for query {i} ...')
            scores.append(self.predict_single(queries[i]))
        return numpy.array(scores, dtype=numpy.float32)

    def vote(self, replace: bool = True) -> numpy.array:
        """ Get ensemble scores with custom voting and normalization

        :param replace: whether to replace internal scores with new scores.
        :return: 1-d array of outlier scores for fitted data.
        """
        scores = self._vote(numpy.stack(self._individual_scores))
        if replace:
            self._scores = scores
        return scores

    def cluster_cardinality(self, graph: Graph) -> Scores:
        """ Determines outlier scores for points by considering the relative cardinalities of the clusters in the graph.

        Points in clusters with relatively low cardinalities are the outliers.

        :param graph: Graph on which to calculate outlier scores.
        :return: A dict of index -> outlier score
        """
        return self._score_points(self._cluster_cardinality(graph))

    def component_cardinality(self, graph: Graph) -> Scores:
        """ Determines outlier scores by considering the relative cardinalities of the connected components of the graph.

        Points in components of relatively low cardinalities are the outliers

        :param graph: Graph on which to calculate outlier scores.
        :return: A dict of index -> outlier score
        """
        return self._score_points(self._component_cardinality(graph))

    def graph_neighborhood(self, graph: Graph, eccentricity_fraction: float = 0.25) -> Scores:
        """ Determines outlier scores by the considering the relative graph-neighborhood of clusters.

        Subsumed clusters are assigned the highest score of all subsuming clusters.
        Points in clusters with relatively small neighborhoods are the outliers.

        :param graph: Graph on which to calculate outlier scores.
        :param eccentricity_fraction: The fraction, in the (0, 1] range, of a cluster's eccentricity for which to compute the size of the neighborhood.
        :return: A dict of index -> outlier score
        """
        return self._score_points(self._graph_neighborhood(graph, eccentricity_fraction))

    def parent_cardinality(self, graph: Graph, weight: typing.Callable[[int], float] = None) -> Scores:
        """ Determines outlier scores for points by considering ratios of cardinalities of parent-child clusters.

        The ratios are weighted by the child's depth in the tree, and are then accumulated for each point in each cluster in the graph.
        Points with relatively high accumulated ratios are the outliers.

        :param graph: Graph on which to calculate outlier scores.
        :param weight: A function for weighing the contribution of each depth.
                       Takes an integer depth and returns a weight.
                       Defaults to 1 / sqrt(depth).
        :return: A dict of index -> outlier score
        """
        return self._score_points(self._parent_cardinality(graph, weight))

    def stationary_probabilities(self, graph: Graph, steps: int = 16) -> Scores:
        """ Compute the Outlier scores based on the convergence of a random walk with weighted edges on each component of the Graph.

        For each component on the graph, compute the convergent transition matrix for that graph.
        Clusters with low values in that matrix are the outliers.

        :param graph: The graph on which to compute outlier scores.
        :param steps: number of steps to wait for convergence.
        :return: A dict of index -> outlier score
        """
        return self._score_points(self._stationary_probabilities(graph, steps))

    def vertex_degree(self, graph: Graph) -> Scores:
        """ Compute the Outlier scores based on the degree of each vertex of the Graph.

        For each cluster in teh graph, its outlier score is inversely proportional to its degree in the graph.

        :param graph: The graph on which to compute outlier scores.
        :return: A dict of index -> outlier score
        """
        return self._score_points(self._vertex_degree(graph))

    def _ensemble(self):
        """ Ensemble of individual methods.
        """
        self._cluster_scores = list()
        argpoints = self._manifolds[0].root.argpoints
        self._individual_scores = list()
        for name, graph in self._graphs:
            method = self._names[name]
            self._cluster_scores.append(method(graph))
            scores: Scores = self._score_points(self._cluster_scores[-1])
            self._individual_scores.append(numpy.asarray([scores[i] for i in argpoints], dtype=float))

        # store scores for fitted data.
        self._scores = self.vote()
        return

    def _vote(self, scores: numpy.array):
        """ Vote among all individual-scores for ensemble-score.

        :param scores: An array of shape (num_graphs, num_points) of scores to be voted among.
        """
        if self._voting_mode == 'mean':
            scores = numpy.mean(scores, axis=0)
        elif self._voting_mode == 'product':
            scores = numpy.product(scores, axis=0)
        elif self._voting_mode == 'median':
            scores = numpy.median(scores, axis=0)
        elif self._voting_mode == 'min':
            scores = numpy.min(scores, axis=0)
        elif self._voting_mode == 'max':
            scores = numpy.max(scores, axis=0)
        elif self._voting_mode == 'p25':
            scores = numpy.percentile(scores, 25, axis=0)
        elif self._voting_mode == 'p75':
            scores = numpy.percentile(scores, 75, axis=0)
        else:
            # TODO: Investigate other voting methods.
            raise NotImplementedError(f'voting mode {self._voting_mode} is not implemented. Try one of {VOTING_MODES}.')

        return scores

    # noinspection PyMethodMayBeStatic
    def _score_points(self, scores: ClusterScores) -> Scores:
        """ Translate scores for clusters into scores for points. """
        scores = {point: float(score) for cluster, score in scores.items() for point in cluster.argpoints}
        for i in self._manifolds[0].root.argpoints:
            if i not in scores:
                # TODO: Rip off this band-aid and investigate source of missing points
                #  So far, this only triggers for exactly one point in exactly one dataset from our suite of datasets.
                #  No idea why.
                scores[i] = 0.5
        return scores

    def _normalize_scores(self, scores: typing.Union[ClusterScores, Scores], high: bool) -> typing.Union[ClusterScores, Scores]:
        """ Normalizes scores to lie in the [0, 1] range.

        :param scores: A dictionary of outlier rankings for clusters.
        :param high: True if high scores denote outliers, False if low scores denote outliers.

        :return: A dict of cluster -> normalized outlier score.
        """
        if self.normalization_mode is None:
            normalized_scores: list[float] = [(s if high is True else (1 - s)) for s in scores.values()]
        else:
            sign = 1 if high is True else -1
            normalized_scores: list[float] = list(helpers.normalize(
                values=sign * numpy.asarray([score for score in scores.values()], dtype=float),
                mode=self.normalization_mode,
            ))

        return {cluster: float(score) for cluster, score in zip(scores.keys(), normalized_scores)}

    def _cluster_cardinality(self, graph: Graph) -> ClusterScores:
        logger.info(f'with metric {graph.metric} on graph {list(graph.depth_range)} with {graph.cardinality} clusters.')
        return self._normalize_scores({cluster: cluster.cardinality for cluster in graph.clusters}, False)

    def _component_cardinality(self, graph: Graph) -> ClusterScores:
        logger.info(f'with metric {graph.metric} on graph {list(graph.depth_range)} with {graph.cardinality} clusters.')
        scores: dict[Cluster, int] = {
            cluster: component.cardinality
            for component in graph.components
            for cluster in component.clusters
        }
        return self._normalize_scores(scores, False)

    # noinspection PyMethodMayBeStatic
    def _inherit_subsumed(self, subsumed_neighbors, scores):
        for master, subsumed in subsumed_neighbors.items():
            for cluster in subsumed:
                if cluster in scores:
                    scores[cluster] = max(scores[cluster], scores[master])
                else:
                    scores[cluster] = scores[master]
        return scores

    def _graph_neighborhood(self, graph: Graph, eccentricity_fraction: float = 0.25) -> ClusterScores:
        logger.info(f'Running method GN with metric {graph.metric} on graph {list(graph.depth_range)} with {graph.cardinality} clusters.')
        if not (0 < eccentricity_fraction <= 1):
            raise ValueError(f'eccentricity fraction must be in the (0, 1] range. Got {eccentricity_fraction:.2f} instead.')

        pruned_graph, subsumed_neighbors = graph.pruned_graph

        def _neighborhood_size(start: Cluster, steps: int) -> int:
            """ Returns the number of clusters within 'steps' of 'start'. """
            visited: set[Cluster] = set()
            frontier: set[Cluster] = {start}
            for _ in range(steps):
                if frontier:
                    visited.update(frontier)
                    frontier = {
                        neighbor for cluster in frontier
                        for neighbor in pruned_graph.neighbors(cluster)
                        if neighbor not in visited
                    }
                else:
                    break
            return len(visited)

        scores: dict[Cluster, int] = {
            cluster: _neighborhood_size(cluster, int(pruned_graph.eccentricity(cluster) * eccentricity_fraction))
            for cluster in pruned_graph.clusters
        }
        scores = self._inherit_subsumed(subsumed_neighbors, scores)
        return self._normalize_scores(scores, False)

    def _parent_cardinality(self, graph: Graph, weight: typing.Callable[[int], float] = None) -> ClusterScores:
        logger.info(f'Running method PC with metric {graph.metric} on graph {list(graph.depth_range)} with {graph.cardinality} clusters.')

        weight = (lambda d: 1 / (d ** 0.5)) if weight is None else weight
        scores: ClusterScores = {cluster: 0 for cluster in graph.clusters}
        for cluster in graph:
            ancestry = graph.manifold.ancestry(cluster)
            for i in range(1, len(ancestry)):
                scores[cluster] += (weight(i) * ancestry[i-1].cardinality / ancestry[i].cardinality)
        return self._normalize_scores(scores, True)

    def _stationary_probabilities(self, graph: Graph, steps: int = 16) -> ClusterScores:
        logger.info(f'Running method SP with metric {graph.metric} on graph {list(graph.depth_range)} with {graph.cardinality} clusters.')
        scores: dict[Cluster, float] = {cluster: -1 for cluster in graph.clusters}
        pruned_graph, subsumed_neighbors = graph.pruned_graph
        for component in pruned_graph.components:
            if component.cardinality > 1:
                clusters, matrix = component.as_matrix
                matrix = matrix / numpy.sum(matrix, axis=1)[:, None]

                for _ in range(steps):
                    # TODO: Go until convergence. For now, matrix ^ (2 ^ 16) ought to be enough.
                    matrix = numpy.linalg.matrix_power(matrix, 2)
                steady = matrix
                scores.update({cluster: score for cluster, score in zip(clusters, numpy.sum(steady, axis=0))})
            else:
                scores.update({cluster: 0 for cluster in component.clusters})
        scores = self._inherit_subsumed(subsumed_neighbors, scores)
        return self._normalize_scores(scores, False)

    def _vertex_degree(self, graph: Graph) -> ClusterScores:
        logger.info(f'Running method VD with metric {graph.metric} on graph {list(graph.depth_range)} with {graph.cardinality} clusters.')
        scores: dict[Cluster, float] = {cluster: graph.vertex_degree(cluster) for cluster in graph.clusters}
        return self._normalize_scores(scores, False)



import pathlib

import toml
from setuptools import setup

with open(pathlib.Path(__file__).parent.joinpath('README.md'), 'r') as f:
    long_description = f.read()

cargo = toml.load('Cargo.toml')
setup(
    name='pyclam',
    version=cargo['package']['version'],
    packages=['pyclam', 'pyclam.anomaly_detection', 'pyclam.classification', 'pyclam.core', 'pyclam.search', 'pyclam.utils'],
    url='https://github.com/URI-ABD/clam',
    license='MIT',
    author='; '.join(cargo['package']['authors']),
    author_email='',
    description='Clustered Learning of Approximate Manifolds',
    long_description=long_description,
    long_description_content_type='text/markdown',
    install_requires=['numpy>=1.22,<1.23', 'scipy>=1.8,<1.9', 'toml>=0.10,<0.11', 'scikit-learn>=1.1,<1.2', 'h5py>=3.7,<3.8'],
    python_requires='>=3.9,<3.11',
)



import logging
import typing

import numpy

from ..anomaly_detection import CHAODA
from ..utils import constants

logger = logging.getLogger(__name__)
logger.setLevel(constants.LOG_LEVEL)


class Classifier:

    def __init__(self, **kwargs):
        """ Creates and initializes a CLAM Classifier.

        See CHAODA for the list of input arguments.
        """
        self.kwargs = kwargs

        self._bowls: typing.Dict[int, CHAODA] = dict()

    def fit(self, data: numpy.ndarray, labels: typing.List[int], *, voting: str = 'mean') -> 'Classifier':
        """ Fits the Classifier to the data.

        Args:
            data: 2d array where the rows are instances and the columns are features.
            labels: List of enumerated labels for each row of data.
            voting: How to vote among scores.

        Returns:
            the fitted Classifier.
        """
        labels = numpy.array(labels)
        unique_labels = numpy.unique(labels)
        for label in unique_labels:
            indices = list(numpy.argwhere(labels == label))

            logger.info(f'Fitting CHAODA object for label {label} ...')
            bowl = CHAODA(**self.kwargs)
            bowl = bowl.fit(data, indices=indices, voting=voting)
            self._bowls[label] = bowl

        return self

    def predict_single(self, query: numpy.ndarray) -> int:
        """ Predicts the label for a single query.
        """
        labels = list(self._bowls.keys())
        scores = numpy.array(list(bowl.predict_single(query) for bowl in self._bowls.values()), dtype=numpy.float32)
        min_index = numpy.argmin(scores)
        return labels[min_index]

    def predict(self, queries: numpy.ndarray) -> typing.List[int]:
        """ Predicts the label for a set of queries.
        """
        labels = list()
        for i in range(queries.shape[0]):
            logger.info(f'Predicting class for query {i} ...')
            labels.append(self.predict_single(queries[i]))
        return labels



import abc
import collections
import heapq
import logging
import typing

import numpy
from scipy.spatial.distance import cdist

from .manifold import Cluster
from .manifold import Manifold
from ..utils import helpers


class Criterion(abc.ABC):
    pass


class ClusterCriterion(Criterion):

    @abc.abstractmethod
    def __call__(self, cluster: Cluster) -> bool:
        pass


class SelectionCriterion(Criterion):

    @abc.abstractmethod
    def __call__(self, root: Cluster) -> set[Cluster]:
        pass


class MaxDepth(ClusterCriterion):
    """ Allows clustering up until the given depth.
    """

    def __init__(self, depth):
        self.depth = depth

    def __call__(self, cluster: Cluster) -> bool:
        return cluster.depth < self.depth


class AddLevels(ClusterCriterion):
    """ Allows clustering up until current.depth + depth.
    """

    def __init__(self, depth):
        self.depth = depth
        self.start = None

    def __call__(self, cluster: Cluster) -> bool:
        if self.start is None:
            self.start = cluster.depth
        return cluster.depth < (self.start + self.depth)


class MinPoints(ClusterCriterion):
    """ Allows clustering up until there are fewer than points.
    """

    def __init__(self, points):
        self.min_points = points

    def __call__(self, cluster: Cluster) -> bool:
        return cluster.cardinality > self.min_points


class MinRadius(ClusterCriterion):
    """ Allows clustering until cluster.radius is less than radius.
    """

    def __init__(self, radius):
        self.radius = radius

    def __call__(self, cluster: Cluster) -> bool:
        return cluster.radius > self.radius


class MedoidNearCentroid(ClusterCriterion):
    def __init__(self):
        return

    def __call__(self, cluster: Cluster) -> bool:
        distance = cdist(numpy.expand_dims(cluster.centroid, 0), numpy.expand_dims(cluster.medoid, 0), cluster.metric)[0][0]
        logging.debug(f'Cluster {str(cluster)} distance: {distance}')
        return any((
            cluster.depth < 1,
            distance > (cluster.radius * 0.1)
        ))


class UniformDistribution(ClusterCriterion):
    def __init__(self):
        return

    def __call__(self, cluster: Cluster) -> bool:
        distances = cdist(numpy.expand_dims(cluster.medoid, 0), cluster.samples)[0] / (cluster.radius + 1e-15)
        logging.debug(f'Cluster: {cluster}. Distances: {distances}')
        freq, bins = numpy.histogram(distances, bins=[i / 10 for i in range(1, 10)])
        ideal = numpy.full_like(freq, distances.shape[0] / bins.shape[0])
        from scipy.stats import wasserstein_distance
        distance = wasserstein_distance(freq, ideal)
        return distance > 0.25


class Leaves(SelectionCriterion):
    """ Selects the leaves for building the graph.
    """
    def __init__(self):
        return

    def __call__(self, root: Cluster) -> set[Cluster]:
        return {cluster for cluster in root.manifold.layers[-1].clusters}


class Layer(SelectionCriterion):
    """ Selects the layer at the specified depth.
    """
    def __init__(self, depth: int):
        if depth < -1:
            raise ValueError(f'expected a \'-1\' or a non-negative depth. got: {depth}')
        self.depth = depth

    def __call__(self, root: Cluster) -> set[Cluster]:
        manifold: Manifold = root.manifold
        if (self.depth == -1) or (manifold.depth <= self.depth):
            return {cluster for cluster in manifold.layers[-1].clusters}
        else:
            return {cluster for cluster in manifold.layers[self.depth].clusters}


class MetaMLSelect(SelectionCriterion):
    """ Uses a decision function from a trained meta-ml model to select best clusters for graph.
    """
    def __init__(self, predict_auc: typing.Callable[[numpy.array], float], min_depth: int = 4):
        """ A Meta-ML model to select Clusters for a Graph.

        :param predict_auc: A function that takes a cluster's ratios and returns a predicted auc for selecting that cluster.
        """
        if min_depth < 1:
            raise ValueError(f'min-depth must be a positive integer.')
        self.min_depth = min_depth

        # TODO: Investigate discount factors to favor high-cardinality clusters
        #  measure performance change vs runtime improvement for individual methods in CHAODA
        # def discount_factor(cluster: Cluster):
        #     num_points = cluster.manifold.root.cardinality
        #     return 1 + ((cluster.cardinality / num_points) / np.sqrt(num_points))

        def ratios(cluster: Cluster):
            return cluster.manifold.cluster_ratios(cluster)

        self.predict_auc: typing.Callable[[Cluster], float] = lambda cluster: -predict_auc(ratios(cluster))

    def __call__(self, root: Cluster) -> set[Cluster]:
        manifold: Manifold = root.manifold

        candidate_clusters = list(filter(lambda c: c.depth >= self.min_depth, manifold.ordered_clusters))
        normalized_auc = list(map(float, helpers.normalize(
            numpy.asarray(list(map(self.predict_auc, candidate_clusters))),
            mode='gaussian',
        )))

        # Turn dict into max priority queue
        items = list(zip(normalized_auc, candidate_clusters))
        heapq.heapify(items)

        selected = set(filter(lambda c: c.depth < self.min_depth and c.is_leaf, manifold.ordered_clusters))
        excluded: set[int] = set(i for cluster in selected for i in cluster.argpoints)
        while len(items) > 0:
            _, cluster = heapq.heappop(items)
            if len(excluded.intersection(set(cluster.argpoints))) > 0:
                continue
            else:
                selected.add(cluster)
                excluded.update(set(cluster.argpoints))

        return selected


class LFDRange(SelectionCriterion):
    """ Selects clusters based on when their ancestry crosses certain thresholds of LFD values.
    This works a bit like a two-state machine.
    """
    def __init__(self, upper: float, lower: float):
        if not (0. < lower <= upper <= 100.):
            raise ValueError(f'LFDRange expected 0 < lower <= upper <= 100 for upper and lower thresholds.'
                             f'Got: lower: {lower:.2f}, and upper: {upper:.2f}')

        self.upper: float = float(upper)
        self.lower: float = float(lower)

    def __call__(self, root: Cluster) -> set[Cluster]:
        upper, lower, grace_depth = self._range(root)
        return self._select(root, upper, lower, grace_depth)

    def _range(self, root: Cluster) -> tuple[float, float, int]:
        manifold = root.manifold
        lfd_range = [], []
        for depth in range(1, manifold.depth):
            if manifold.layers[depth + 1].cardinality < 2 ** (depth + 1):
                clusters: list[Cluster] = [cluster for cluster in manifold.layers[depth] if cluster.cardinality > 2]
                if len(clusters) > 0:
                    lfds = numpy.percentile(
                        a=[c.local_fractal_dimension for c in clusters],
                        q=[self.upper, self.lower],
                    )
                    lfd_range[0].append(lfds[0]), lfd_range[1].append(lfds[1])
        if len(lfd_range[0]) > 0:
            upper: float = float(numpy.median(lfd_range[0]))
            lower: float = float(numpy.median(lfd_range[1]))
            depth: int = manifold.depth - len(lfd_range[0])
            return upper, lower, depth
        else:
            lfds = numpy.percentile(
                a=[cluster.local_fractal_dimension for cluster in manifold.layers[-1]],
                q=[self.upper, self.lower],
            )
            return float(lfds[0]), float(lfds[1]), manifold.depth
        pass

    @staticmethod
    def _select(root: Cluster, upper: float, lower: float, grace_depth: int) -> set[Cluster]:
        # childless clusters at or before grace depth are automatically selected
        selected: set[Cluster] = {
            cluster
            for cluster in root.manifold.layers[grace_depth].clusters
            if not cluster.children
        }
        inactive: set[Cluster] = {
            cluster
            for cluster in root.manifold.layers[grace_depth].clusters
            if cluster.children
        }
        active: set[Cluster] = set()

        while active or inactive:
            # Select childless clusters from inactive set
            childless: set[Cluster] = {cluster for cluster in inactive if len(cluster.children) == 0}
            # Select childless clusters from active set
            childless.update({cluster for cluster in active if len(cluster.children) == 0})
            selected.update(childless)
            inactive -= childless
            active -= childless

            # Select active clusters that fall below the lower threshold
            selections: set[Cluster] = {cluster for cluster in active if cluster.local_fractal_dimension <= lower}
            selected.update(selections)
            active -= selections

            # activate branches that rise above the upper threshold
            activations: set[Cluster] = {cluster for cluster in inactive if cluster.local_fractal_dimension >= upper}
            active.update(activations)
            inactive -= activations

            # replace active and inactive sets with child clusters
            active = {child for cluster in active for child in cluster.children}
            inactive = {child for cluster in inactive for child in cluster.children}

        return selected


class PropertyThreshold(SelectionCriterion):
    """ Selects clusters when the given property crosses the given percentile.

    If mode is 'above', the cluster qualifies if the property goes above the percentile,
    If mode is 'below', the cluster qualifies if the property goes below the percentile.

    During a BFT of the tree, keep track of the given cluster property.
    Select a cluster only if:
        * it is a leaf, or
        * it qualifies by 'mode' as above, or
        * all children qualify by 'mode'.

    :param value: The cluster property to keep track of. Must be one of 'cardinality', 'radius', or 'lfd'.
    :param percentile: the percentile, in the (0, 100) range, to be crossed.
    :param mode: select when 'above' or 'below' the percentile.
    """
    def __init__(self, value: str, percentile: float, mode: str):
        # TODO: Use Literal for value and mode.
        if value == 'cardinality':
            self.value = lambda c: c.cardinality
        elif value == 'radius':
            self.value = lambda c: c.radius
        elif value == 'lfd':
            self.value = lambda c: c.local_fractal_dimension
        else:
            raise ValueError(f'value must be one of \'cardinality\', \'radius\', or \'lfd\'. Got {value} instead.')

        if 0 < percentile < 100:
            self.percentile: float = percentile
        else:
            raise ValueError(f'percentile must be in the (0, 100) range. Gor {percentile:.2f} instead.')

        if mode == 'above':
            self.mode = lambda c, v: self.value(c) > v
        elif mode == 'below':
            self.mode = lambda c, v: self.value(c) < v
        else:
            raise ValueError(f'mode must be \'above\' or \'below\'. Got {mode} instead.')

    def __call__(self, root: Cluster) -> set[Cluster]:
        threshold = float(numpy.percentile(
            [self.value(cluster) for cluster in root.descendents if cluster.cardinality > 1],
            self.percentile,
        ))
        selected: set[Cluster] = set()
        frontier: set[Cluster] = {root}
        while frontier:
            cluster = frontier.pop()
            if (
                    (cluster.children is None)  # select the leaves
                    or self.mode(cluster, threshold)  # if cluster qualifies
                    or all((map(lambda c: self.mode(c, threshold), cluster.children)))  # if both children qualify
            ):
                selected.add(cluster)
            else:  # add children to frontier
                frontier.update(cluster.children)
        return selected


class Labels(SelectionCriterion):
    """ Uses a given dictionary of index -> label to select optimal clusters.

    During a BFT of the tree, when a cluster has an overwhelming majority of points with the same label,
    that cluster is selected.
    If some data points are unlabelled, then when a cluster contains no labelled points, it too is selected.

    Let 'f' be the fraction of labels in the smallest class in the data,
    and let 't' be a given threshold in the (0, 1] range.
    Then the 'overwhelming majority' is defined as (1 - f * t).
    """
    def __init__(self, labels: dict[int, typing.Any], threshold: float = 0.1):
        """ Instantiates the Labels Selector.

        :param labels: A dictionary of index -> label for som or all of the points in the data.
        :param threshold: A threshold, in the (0, 1] range for determining the size of the 'overwhelming majority.'

        """
        if not (0 < threshold <= 1):
            ValueError(f'threshold must be in the (0, 1] range.')
        self.labels: dict[int, typing.Any] = labels
        self.min_fraction: float = min(dict(collections.Counter(labels.values())).values()) / len(labels)
        self.threshold: float = 1 - self.min_fraction * threshold

    def __call__(self, root: Cluster) -> set[Cluster]:
        selected: set[Cluster] = set()
        frontier: set[Cluster] = {root}
        while frontier:
            cluster = frontier.pop()
            if cluster.children is None:
                selected.add(cluster)
            else:
                labels: list[typing.Any] = [self.labels[p] for p in cluster.argpoints if p in self.labels]
                if len(labels) > 0:
                    max_fraction = max(dict(collections.Counter(labels)).values()) / cluster.cardinality
                    if max_fraction > self.threshold:
                        selected.add(cluster)
                    else:
                        frontier.update(cluster.children)
                else:
                    # TODO: Should we revert to some other selection criterion here?
                    selected.add(cluster)
        return selected
